import functools
from typing import Optional, Iterable

from gelidum import freeze


def freeze_params(params: Optional[Iterable[str]] = None):
    def inner_freeze_params(func):
        """Freeze all input params of a method"""
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_args = tuple([freeze(arg, inplace=False)
                               for arg in args])
            func_kwargs = {
                kwarg_name: freeze(kwarg, inplace=False)
                if kwarg_name in params else kwarg
                for kwarg_name, kwarg in kwargs.items()
            }
            return func(*func_args, **func_kwargs)
        return wrapper
    return inner_freeze_params



import copy
import sys
from typing import List, Tuple, Set, Dict, Any

from frozendict import frozendict

from gelidum.frozen import make_frozen_class
from gelidum.utils import isbuiltin


def freeze(obj: Any, inplace: bool = False) -> Any:
    if isbuiltin(obj):
        return obj

    class_name = type(obj).__name__
    freeze_func_name = f"__freeze_{class_name}"
    this_module = sys.modules[__name__]
    if hasattr(this_module, freeze_func_name):
        freeze_func = getattr(this_module, freeze_func_name)
        return freeze_func(obj, inplace=inplace)

    if isinstance(obj, object):
        return __freeze_object(obj, inplace=inplace)

    raise ValueError(f"object of type {obj.__class__} not frozen")


def __freeze_bytearray(obj: bytearray, inplace: bool = False) -> bytes:
    return bytes(obj)


def __freeze_dict(obj: Dict, inplace: bool = False) -> frozendict:
    return frozendict({key: freeze(value, inplace=inplace)
                       for key, value in obj.items()})


def __freeze_list(obj: List, inplace: bool = False) -> Tuple:
    return tuple(freeze(item, inplace=inplace) for item in obj)


def __freeze_tuple(obj: Tuple, inplace: bool = False) -> Tuple:
    return tuple(freeze(item, inplace=inplace) for item in obj)


def __freeze_set(obj: Set, inplace: bool = False) -> frozenset:
    return frozenset([freeze(item, inplace=inplace) for item in obj])


def __freeze_object(obj: object, inplace: bool = False) -> object:
    if inplace:
        frozen_obj = obj
    else:
        frozen_obj = copy.deepcopy(obj)
    for attr, value in frozen_obj.__dict__.items():
        attr_value = getattr(frozen_obj, attr)
        setattr(frozen_obj, attr, freeze(attr_value, inplace=inplace))
    frozen_obj.__class__ = make_frozen_class(klass=obj.__class__, attrs=list(obj.__dict__.keys()))
    return frozen_obj



from typing import Type, List
from gelidum.exceptions import FrozenException


class FrozenBase(object):
    def __setattr__(self, key, value):
        raise FrozenException(f"Can't assign '{key}' on immutable instance")

    def __set__(self, *args, **kwargs):
        raise FrozenException("Can't assign setter on immutable instance")

    def __delattr__(self, name):
        raise FrozenException(
            f"Can't delete attribute '{name}' on immutable instance")

    def __setitem__(self, key, value):
        raise FrozenException("Can't set key on immutable instance")

    def __delitem__(self, key):
        raise FrozenException("Can't delete key on immutable instance")

    def __reversed__(self):
        raise FrozenException("Can't reverse on immutable instance")


def make_frozen_class(klass: Type[object], attrs: List[str]):
    frozen_class = type(
        f"Frozen{klass.__name__}",
        (klass, FrozenBase),
        {
            "__slots__": tuple(),
            **{attr: None for attr in attrs}
        }
    )
    return frozen_class



import os
from setuptools import setup

root_dir_path = os.path.dirname(os.path.abspath(__file__))

long_description = open(os.path.join(root_dir_path, "README.md")).read()

requirements_path = os.path.join(root_dir_path, "requirements.txt")
with open(requirements_path) as requirements_file:
    requirements = requirements_file.readlines()

setup(
    name="gelidum",
    version="0.3",
    author="Diego J. Romero López",
    author_email="diegojromerolopez@gmail.com",
    description="Freeze your python objects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=requirements,
    license="MIT",
    keywords="freeze python object",
    url="https://github.com/diegojromerolopez/gelidum",
    packages=["gelidum"],
    package_dir={"gelidum": "gelidum"},
    data_files=[],
    include_package_data=True,
    scripts=[]
)
